# -*- coding: iso-8859-1 -*-
import sys
from optparse import OptionParser

import RelaxNG as rng

from ll.xist import xnd, xsc, sims, parsers, converters
from ll.xist.ns import meta, xml, html, htmlspecials

f = file("log.txt", 'w')

class R2X:
	def __init__(self, **args):
		self.args = args
		# evtl. self.__dict__.update(args)
		
		# name: keine vorinitialisierung
		# modeltype: keine vorinitialisierung
		
		# empty
		if not self.args.has_key("empty"):
			self.args["empty"] = False
		
		# text
		if not self.args.has_key("text"):
			self.args["text"] = False

		# modelargs: list element-names
		if not self.args.has_key("modelargs"):
			self.args["modelargs"] = []
		
		# attrs: list of xnd.attr-objects
		if not self.args.has_key("attrs"):
			self.args["attrs"] = []
		
		# refs: list of references
		if not self.args.has_key("refs"):
			self.args["refs"] = []

	def construct(self, etree, layer, enter=True):
		"""
		etree: xist subtree starting with element tag
		return onedimensional list of R2X object: [self, children] * if children don't contain description for self
		"""
		
		"""
		@EP
			anyname
			except
		"""
		# list of element-children
		L = [self]
		
		# name
		if not self.args.has_key("name"):
			self.args["name"] = str(etree.attrs["name"])
		
		# attrs
		self.args["attrs"] = attrget(etree, self.args["attrs"])
		
		# text
		if not self.args["text"]:
			self.args["text"] = testtext(etree)

		# ref
		refwalk = etree.walknode(EWalk(rng.ref, (rng.grammar, rng.element_), 1))
		for r in refwalk:
			if r not in self.args["refs"]:
				self.args["refs"].append("%s:%s"%(layer, str(r.attrs["name"])))

		# parentref
		prefwalk = etree.walknode(EWalk(rng.parentRef, (rng.grammar, rng.element_), 1))
		for p in prefwalk:
			if p not in self.args["refs"]:
				self.args["refs"].append("%s:%s"%(layer-1, str(p.attrs["name"])))
		
		# empty
		if not self.args["empty"]:
			emptywalk = etree.walknode(EWalk(rng.empty, (rng.grammar, rng.element_), 1))
			try:
				emptywalk.next()
				self.args["empty"] = True
			except StopIteration:
				pass
		
		# start
		startwalk = etree.walknode(EWalk(rng.start, rng.element_, 1)) 
		try:
			start = startwalk.next()
			s = R2X(name = self.args["name"])
			s.construct(start, layer+1, enter=False)
			x = self.merge([s])
			self = x[0]
		except StopIteration:
			pass
		
		# modelargs + rekursiver Aufruf
		elems = etree.walknode(EWalk(rng.element_, rng.grammar, 1))
		for elem in elems:
			# build subtree
			if enter:
				r2x = R2X()
				l = r2x.construct(elem, layer)
				# build modelargs
				if r2x.args["name"] not in self.args["modelargs"]:
					self.args["modelargs"].append(r2x.args["name"])
				# merge
				L = self.merge(l, L)
			else:
				if str(elem.attrs["name"]) not in self.args["modelargs"]:
					self.args["modelargs"].append(str(elem.attrs["name"]))
		
		if enter:
			return L
		
	def merge(self, li, l=[]):
		"""
		method to merge this element in a list, optional with another list
		li: a list of Element instance
		return list with element and (2nd list) merged in
		"""
		L = [self]
		L.extend(li)
		L.extend(l)
		li = []
		for i in L:
			found = False
			for j in li:
				if j.args["name"] == i.args["name"]:
					found = True
					# merge modelargs
					for ma in i.args["modelargs"]:
						if ma not in j.args["modelargs"]:
							j.args["modelargs"].append(ma)
					# merge attrs
					j.args["attrs"] = attrmerge(i.args["attrs"], j.args["attrs"])
					# merge text
					if not j.args["text"]:
						j.args["text"] = i.args["text"]
					# merge refs
					for re in i.args["refs"]:
						if re not in j.args["refs"]:
							j.args["refs"].append(re)
			if not found:
				li.append(i)
		return li

	def refmerge(self, refdict):
		"""
		refdict: a dictionary of this form refdict = { "layer:refname" : [[first layer nodes], Text]}
		if one of this node's direct children is <ref>, the direct children of the corresponding <define> must be added to modelargs
		raises Exception if there's no corresponding <define>
		"""
		for ref in self.args["refs"]:
			# corresponding ref/define check
			try:
				content = refdict[ref]
			except KeyError:
				raise "Reference error, %s can't be found"%ref
			
			content.args["name"] = self.args["name"]
			l = self.merge([content])
			self = l[0]
	
	def modelmake(self):
		"""
		sets value for modeltype
		"""
		# Any
		#@TODO: existiert noch nicht, nur bei anyname
		
		if self.args["text"]:
			if self.args["modelargs"]:
				# ElementsOrText
				self.args["modeltype"] = "sims.ElementsOrText"
			else:
				# NoElements
				self.args["modeltype"] = "sims.NoElements"
		else:
			if self.args["modelargs"]:
				# Elements
				self.args["modeltype"] = "sims.Elements"
			else:
				pass
				if self.args["empty"]:
					# Empty
					self.args["modeltype"] = "sims.Empty"
				else:
					# NoElementsOrText
					self.args["modeltype"] = "sims.NoElementsOrText"
	
	def __repr__(self):
		"""
		Method for displaying R2X-Objects
		"""
		attrs = "\n  Attrs: "
		for a in self.args["attrs"]:
			attrs = attrs + a.name + ", "
		if attrs == "\n  Attrs: ":
			attrs = ""
			
		modelargs = "\n  Modelargs: "
		for m in self.args["modelargs"]:
			modelargs = modelargs + m + ", "
		if modelargs == "\n  Modelargs: ":
			modelargs = ""
			
		refs = "\n  Refs: "
		for m in self.args["refs"]:
			refs = refs + m + ", "
		if refs == "\n  Refs: ":
			refs = ""
			
		sonst = "\n  Sonstiges: "
		if self.args["text"]:
			sonst = sonst + " Text"
		if self.args["empty"]:
			sonst = sonst + " Empty"
		if self.args.has_key("modeltype"):
			sonst = sonst + " " + self.args["modeltype"]
		if sonst == "\n  Sonstiges: ":
			sonst = ""
		
		return "\n\n- %s at 0x%x%s%s%s%s" % ( self.args["name"], id(self), attrs, modelargs, refs, sonst)

def testtext(etree):
	"""
	testtext
	function to test if any text node is in an element
	"""
	text = False
	# text, mixed, data, list, value
	try:
		etree.walknode(EWalk((rng.text, rng.mixed, rng.data, rng.list, rng.value), (rng.attribute, rng.element_, rng.grammar), 1)).next()
		text = True
	except StopIteration:
		pass
	return text

def attrget(etree, attrlist=[]):
	"""
	attrget
	function to get a list of attributes of etree
	"""
	# required attrs
	attribs = etree[rng.attribute]
	for a in attribs:
		#@TODO: type feststellen
		t = xnd.Attr(str(a.attrs["name"]), "xsc.TextAttr") 
		t.required="True"
		#@TODO: andere eigenschaften?
		attrlist = attrmerge(attrlist, [t])
	
	# not required attrs 
	attribs = etree.walknode(EWalk(rng.attribute, (rng.element_, rng.grammar), 1))
	for a in attribs:
		found = False
		for attr in attrlist:
			if attr.name == str(a.attrs["name"]):
				found = True
		if not found:
			#@TODO: type feststellen
			t = xnd.Attr(str(a.attrs["name"]), "xsc.TextAttr")
			#@TODO: andere eigenschaften?
			t.required = False
			attrlist = attrmerge(attrlist, [t])

	return attrlist

def attrmerge(attlist, alist):
	"""
	function to merge two lists of xnd.attribute - objects
	"""
	alist.extend(attlist)
	attlist = []
	for a in alist:
		found = False
		for att in attlist:
			if att.name == a.name:
				found = True
				# merge required
				if att.required:
					att.required = a.required
				# @TODO: merge type / other attributes
		if not found:
			attlist.append(a)
	return attlist

class EWalk(object):
	"""
	Extended Walk traversal filter
	
	<dlist>
	<term>tree</term><item> </item>
	<term>ns_a</term><item> </item>
	<term>ns_b</term><item> </item>
	</dlist>
	"""
	def __init__(self, ns_a, ns_b=None, count=0):
		self.ns_a = ns_a
		self.ns_b = ns_b
		self.count = count
		
	def __call__(self, cursor):
		if self.count!=0:
			self.count = self.count-1
			return (xsc.entercontent,)
			
		# wenn Knoten vom Typ b: not yield, not content
		if self.ns_b is not None:
			if isinstance(cursor.node, self.ns_b):
				return (False,)
	
		# wenn Knoten vom Typ a: yield, not content
		if isinstance(cursor.node, self.ns_a):
			return (True,)

		# beides nicht: entercontent
		return (xsc.entercontent,)


def define(refdict, diter, layer=0):
	"""
	function to build refdict out of an iterator of define-elements
	"""
	for item in diter:
		key = "%s:%s"%(layer, str(item.attrs["name"]))
		if refdict.has_key(key):
			rx = refdict[key]
		else:
			rx = R2X()
		rx.construct(item, layer, enter=False)
		
		refdict[key] = rx
	return refdict
	
def build(gramtree, l, refdict={}, layer=0):
	# recursive
	for gram in gramtree.walknode(EWalk(rng.grammar, count=1)):
		l, refdict = build(gram, l, refdict, layer+1)
	
	diter = gramtree.walknode(EWalk(rng.define, rng.grammar, 1))
	refdict = define(refdict, diter, layer)

	elem = gramtree.walknode(EWalk(rng.element_, rng.grammar, 1))
	
	for e in elem:
		rx = R2X()
		li = rx.construct(e, layer)
		l = rx.merge(li, l)
	
	return l, refdict

def incl(node, converter):
	if isinstance(node, rng.include):
		# alle define, start raussuchen, speichern
		
		# @TODO: sinnvollere Lsung
		f = str(node.attrs["href"])
		f = f[5:]
		# parse file
		node = parsers.parseFile(f)
		print "bla"
	return node

def external(node, converter):
	if isinstance(node, rng.externalRef):
		#@TODO: sinnvollere Lsung
		f = str(node.attrs["href"])
		f = f[5:]
		# parse file
		node = parsers.parseFile(f)
	return node

if __name__=="__main__":
# @TODO: namespace: optparse? Oder auch per sys.argv?
# @TODO: Zielfile, NS-Name
	print "reading filename...\n"
	
	if len(sys.argv)!=2:
		raise "Please define a source file as first argument"
	else:
		filename = sys.argv[1]
		
	print "parsing file...\n"
	
	prefixes = xsc.Prefixes([rng], xml=xml)
	try:
		tree = parsers.parseFile(filename, prefixes=prefixes)
	except IOError:
		raise "Please define a source file as first argument"
	
	#@TODO: mapped
	tree = tree.mapped(incl)
	tree = tree.mapped(external)

	try:
		root = tree[rng.grammar].next()
	except StopIteration:
		try:
			root = tree.walknode(xsc.FindTypTop(rng.element_)).next()
		except:
			raise "root-element error"
	
	# refdict = { "layer:refname" : [[element nodes], text, [attribs]] }
	# l: list of found elements
	l, refdict = build(root, [])
	f.write(refdict.__repr__())
	f.write("\n")
	f.write(l.__repr__())
	f.write("\n")

	# Namespace
	f.write("building namespace...\n\n")
	ns = xnd.Namespace("blafoo")

	for r2x in l:
		f.write(r2x.__repr__())
		r2x.refmerge(refdict)
		f.write(r2x.__repr__())
		r2x.modelmake()
		f.write(r2x.__repr__())
		# name, modeltype, modelargs
		e = xnd.Element(r2x.args["name"])
		e.modeltype = r2x.args["modeltype"]
		e.modelargs = r2x.args["modelargs"]
		# attrs
		for i in r2x.args["attrs"]:
			# already xnd.attr-objects
			e.attrs.append(i)
		ns.content.append(e)
		
	file("ns.py", "w").write(ns.aspy())
